/* $Id$ */

/*
 * $Log$
 */


#include <cfg/os.h>

#include <string.h>
#include <io.h>

#include <dev/board.h>
#include <dev/urom.h>

#include <sys/thread.h>
#include <sys/timer.h>
#include <sys/heap.h>
#include <sys/confnet.h>
#include <sys/socket.h>

#include <arpa/inet.h>
#include <net/route.h>

#include <pro/httpd.h>
#include <pro/wins.h>
#include <pro/dhcp.h>

#include "boothttp.h"
#include "flashCgi.h"

extern int g_bootImage;
static int l_autoBootImage;

#define UNUSED_VAR(_x)  _x=_x

void bootImage(void) {

    NutSleep(500);
    asm volatile ("@ Start Application" "\n\t"
                  "ldr r0, =" IMAGE_START_ADDRESS "\n\t"
                  "bx  r0" "\n\t"
                  :::"r0"
    );
}


THREAD(WINS, arg) {

    UNUSED_VAR(arg);
    NutWinsNameQuery(WINS_NAME,  confnet.cdn_ip_addr);
    for(;;)
        NutSleep(AUTOBOOT_TIME);
}

THREAD(HTTPD, arg) {

    UNUSED_VAR(arg);

    TCPSOCKET *sock;
    FILE *stream;

    for (;;) {
        if ((sock = NutTcpCreateSocket()) == 0) {
            printf("Create Socket failed mem=%d.\n", (int)(NutHeapAvailable()));
            NutSleep(5000);
            continue;
        }

        NutTcpAccept(sock, HTTPD_PORT);
        printf("WWW Accept\n");
        l_autoBootImage = 0;

        while (NutHeapAvailable() < LOW_MEM_MARK) {
            printf("Low mem %d", (int)(NutHeapAvailable()));
            NutSleep(1000);
        }
        if ((stream = _fdopen((int) ((uptr_t) sock), "r+b")) == 0) {
            printf("Creating stream device failed.\n");
        } else {
            NutHttpProcessRequest(stream);
            fclose(stream);
        }
        NutTcpCloseSocket(sock);
        if (g_bootImage) {
            bootImage();
            g_bootImage = 0;
        }
    }
}

int initNetwork(void) {

    if(NutRegisterDevice(&DEV_ETHER, 0x8300, 0)) {
        return -1;
    }

    /* use DHCP */
    if ((NutDhcpIfConfig("eth0", 0, 60000)) && (NutDhcpIfConfig("eth0", MAC, 60000))) {
        /* use EEPROM */
        if (NutNetLoadConfig("eth0") == -1) {
            u_long ip_addr = inet_addr((const char *)(MYIP));
            u_long ip_mask = inet_addr((const char *)(MYMASK));
            memcpy(confnet.cdn_mac, MAC, sizeof(confnet.cdn_mac));
            confnet.cdn_ip_addr = ip_addr;
            confnet.cdn_ip_mask = ip_mask;
        }
        NutNetIfConfig("eth0", confnet.cdn_mac, confnet.cdn_ip_addr, confnet.cdn_ip_mask);
    }

    NutThreadCreate(THREAD_WINS, WINS, NULL, WINS_THREAD_STACK);

    printf("IP-Address: %s.\n", inet_ntoa(confnet.cdn_ip_addr));
    printf("Netmask   : %s.\n", inet_ntoa(confnet.cdn_ip_mask));

    if ((confnet.cdn_ip_addr) && (confnet.cdn_gateway)) {
        NutIpRouteAdd(0, 0, confnet.cdn_gateway, &DEV_ETHER);
        printf("IP gate   : %s.\n", inet_ntoa(confnet.cdn_gateway));
    }
    return 0;
}


int main(void) {


    int i;
    u_long baud = 115200;
    NutRegisterDevice(&devDebug0, 0, 0);
    freopen("uart0", "w", stdout);
    _ioctl(_fileno(stdout), UART_SETSPEED, &baud);
    NutSleep(200);

    initNetwork();
    NutRegisterDevice(&devUrom, 0, 0);

    g_bootImage = 0;
    l_autoBootImage = 1;


    /* thread to get dynamically IP settings */
    // NutSetupStart();


    if (NutRegisterHttpRoot(HTTPD_ROOT)) {
        printf("Error register root.\n");
    }

    NutRegisterCgi(CGI_FLASH, FlashCgi);
    NutRegisterCgi(CGI_STARTIMAGE, StartImageCgi);
    NutRegisterCgi(CGI_STATUS, StatusFlashCgi);

    for (i = 1; i <= 2; i++) {
        THREAD_HTTPD[strlen(THREAD_HTTPD)] = '0' + i;
        NutThreadCreate(THREAD_HTTPD, HTTPD, (void *) (uptr_t) i, HTTPD_THREAD_STACK);
    }

    for (;;) {
        NutSleep(AUTOBOOT_TIME);
        if (l_autoBootImage) {
            printf("Auto start image.\n");
            bootImage();
            l_autoBootImage = 0;
        }
    }

    return 0;

}

