/* $Id$ */

/*
 * $Log$
 */

#include <string.h>
#include <io.h>

#include <sys/heap.h>
#include <sys/timer.h>
#include <pro/httpd.h>

#include "flash.h"
#include "flashCgi.h"
#include "flashCgiTxt.h"

typedef enum  {
    NOERROR                = 0,
    ERROR_UNKNOWN          = -1,
    ERROR_IMAGE_EOF        = -99,
    ERROR_REQUEST          = -10,
    ERROR_NOBAUNDARY       = -11,
    ERROR_WRONGBAUNDARY    = -12,
    ERROR_TOMANYSECTORS    = -20,
    ERROR_WRITEFLASH       = -21,
    ERROR_ERASESECTOR      = -22,
    ERROR_FLASHLOCKED      = -30
} error_t;

typedef enum  {
    STATUS_UNKNOWN         = -1,
    STATUS_NOTSTARTED      = 0,
    STATUS_READY           = 99,
    STATUS_FLASHREQUEST    = 10,
    STATUS_READIMAGE       = 30,
    STATUS_ERASESECTORS    = 40,
    STATUS_WRITEIMAGE      = 50,
    STATUS_IMAGEWRITTEN    = 60,
    STATUS_START_IMAGE     = 70
} status_t;

// globals
int g_bootImage;

// locals
static error_t  l_flashError = NOERROR;
static status_t l_flashStatus = STATUS_NOTSTARTED;
static int l_lockFlashCgi;

static char *html_mt = "text/html";

error_t eraseFlash(long sumBytes) {

    int ret;
    int erasedSectors;
    int sumSectors = (sumBytes + SECTORSIZE - 1) / SECTORSIZE;

    if (sumSectors > MAXSECTORS)
        return ERROR_TOMANYSECTORS;

    for (erasedSectors = 0; erasedSectors < sumSectors; erasedSectors++) {
        printf("Erase sector %d.\n", erasedSectors);
        ret = FlashEraseSector((flashptr_t)BASEADDRESS, (flashptr_t)(IMAGEADDRESS + erasedSectors * SECTORSIZE));
        if (ret != 0)
            return ERROR_ERASESECTOR;
    }
    return NOERROR;
}

error_t readWriteSectorsFromStream(FILE *stream, long sumBytes) {

    long bytesFlashed;
    unsigned long data;

    for (bytesFlashed = 0; bytesFlashed < sumBytes; bytesFlashed += 4) {
        if ((bytesFlashed % 400) == 0) {
            if ((bytesFlashed % 12000) == 0) {
                printf("\nWritten %d of %d.", (int) bytesFlashed, (int) sumBytes);
            }
            printf(".");
        }
        if (fread(&data, 1, 4, stream) != 4)
            return ERROR_IMAGE_EOF;
        if (FlashWrite((flashptr_t)BASEADDRESS, (flashptr_t)(IMAGEADDRESS + bytesFlashed), (flashdat_t)data)) {
            return ERROR_WRITEFLASH;
        }
        data >>= 16;
        if (FlashWrite((flashptr_t)BASEADDRESS, (flashptr_t)(IMAGEADDRESS + bytesFlashed + 2), (flashdat_t)data)) {
           return ERROR_WRITEFLASH;
        }
    }
    return NOERROR;
}

int getLine(FILE * stream) {

    int len = 0;
    char data[1];

    while (fread(data, 1, 1, stream) == 1 && *data != 0x0a)
        len++;
    return (len + 1);
}


int outCgi(FILE * stream, REQUEST * req) {

    char *statusTxt;
    char *errorTxt;

    switch  (l_flashStatus) {
        case STATUS_READY:
            statusTxt = STATUS_READY_TXT;
        break;
        case STATUS_NOTSTARTED:
            statusTxt = STATUS_NOTSTARTED_TXT;
        break;
        case STATUS_FLASHREQUEST:
            statusTxt = STATUS_FLASHREQUEST_TXT;
        break;
        case STATUS_READIMAGE:
            statusTxt = STATUS_READIMAGE_TXT;
        break;
        case STATUS_ERASESECTORS:
            statusTxt = STATUS_ERASESECTORS_TXT;
        break;
        case STATUS_WRITEIMAGE:
            statusTxt = STATUS_WRITEIMAGE_TXT;
        break;
        case STATUS_IMAGEWRITTEN:
            statusTxt = STATUS_IMAGEWRITTEN_TXT;
        break;
        case STATUS_START_IMAGE:
            statusTxt = STATUS_START_IMAGE_TXT;
        break;
        default:
            statusTxt = STATUS_UNKNOWN_TXT;
    }

    switch  (l_flashError) {
        case NOERROR:
            errorTxt = NOERROR_TXT;
        break;
        case ERROR_IMAGE_EOF:
            errorTxt = ERROR_IMAGE_EOF_TXT;
        break;
        case ERROR_REQUEST:
            errorTxt = ERROR_REQUEST_TXT;
        break;
        case ERROR_NOBAUNDARY:
            errorTxt = ERROR_NOBAUNDARY_TXT;
        break;
        case ERROR_WRONGBAUNDARY:
            errorTxt = ERROR_WRONGBAUNDARY_TXT;
        break;
        case ERROR_TOMANYSECTORS:
            errorTxt = ERROR_TOMANYSECTORS_TXT;
        break;
        case ERROR_WRITEFLASH:
            errorTxt = ERROR_WRITEFLASH_TXT;
        break;
        case ERROR_ERASESECTOR:
            errorTxt =ERROR_ERASESECTOR_TXT;
        break;
        case ERROR_FLASHLOCKED:
            errorTxt = ERROR_FLASHLOCKED_TXT;
        break;
        case ERROR_UNKNOWN:
            errorTxt = ERROR_UNKNOWN_TXT;
        break;


    }
    fprintf_P(stream, "<p class=\"status\">%s</p>\n", statusTxt);

    if (l_flashError == NOERROR)
        fprintf_P(stream, "<p class=\"status\">%s</p>\n", NOERROR_TXT);
    else
        fprintf_P(stream, "<p class=\"error\">%s</p>\n", errorTxt);
    fputs_P(html_bot, stream);
    fflush(stream);
    return 0;
}


int StartImageCgi(FILE * stream, REQUEST * req) {

    NutHttpSendHeaderTop(stream, req, 200, "Ok");
    NutHttpSendHeaderBot(stream, html_mt, -1);
    printf("Start image.\n");

    fputs_P(html_top, stream);

    if (!l_lockFlashCgi) {
        g_bootImage = 1;
        l_flashStatus = STATUS_START_IMAGE;
    }
    else {
        l_flashError = ERROR_FLASHLOCKED;
    }
    return outCgi(stream, req);
}

int StatusFlashCgi(FILE * stream, REQUEST * req) {

    NutHttpSendHeaderTop(stream, req, 200, "Ok");
    NutHttpSendHeaderBot(stream, html_mt, -1);
    fputs_P(html_top, stream);
    printf("Get status.\n");
    return outCgi(stream, req);
}

int FlashCgi(FILE * stream, REQUEST * req)
{

    char *baundary;
    int lengthBaundary;
    int checkBaundary;
    char data[2];
    int lenContentDisposition;
    int lenContentType;
    long sumBytes;

    NutHttpSendHeaderTop(stream, req, 200, "Ok");
    NutHttpSendHeaderBot(stream, html_mt, -1);
    fputs_P(html_top, stream);

    printf("Flash CGI\n");
    if (l_lockFlashCgi) {
        printf("Flash CGI locked\n");
        return outCgi(stream, req);
    }

    l_flashError = NOERROR;
    l_flashStatus = STATUS_FLASHREQUEST;

    if (req->req_method != METHOD_POST || !req->req_type) {
        printf("Flash CGI ERROR: bad request\n");
        l_flashError = ERROR_REQUEST;
        return outCgi(stream, req);
    }

    l_flashStatus = STATUS_READIMAGE;

    printf("Flash CGI read image\n");
    baundary = strstr(req->req_type, "boundary=");
    if (baundary == NULL) {
        printf("Flash CGI ERROR: no baundary found\n");
        l_flashError = ERROR_NOBAUNDARY;
        return outCgi(stream, req);
    }

    baundary += 9;
    lengthBaundary = strlen(baundary);
    checkBaundary = getLine(stream) - 4;
    if (checkBaundary != lengthBaundary) {
        printf("Flash CGI ERROR: wrong baundary\n");
        l_flashError = ERROR_WRONGBAUNDARY;
        return outCgi(stream, req);
    }
    printf("Flash CGI baundary length = %d.\n", lengthBaundary);
    lenContentDisposition = getLine(stream);
    lenContentType = getLine(stream);

    if (fread(data, 1, 2, stream) != 2) {
        printf("Flash CGI ERROR: wrong baundary end\n");
        l_flashError = ERROR_WRONGBAUNDARY;
        return outCgi(stream, req);
    }

    sumBytes = req->req_length - lengthBaundary - 2 - lenContentDisposition - lenContentType - lengthBaundary - 2 - 2 - 4 - 4;
    printf("Flash CGI image size = %ld.\n", sumBytes);

    l_lockFlashCgi = 1;

    printf("Flash CGI ID.\n");
    l_flashStatus = FlashIdentifier((flashptr_t)BASEADDRESS);
    printf("I=%d.\n",  l_flashStatus);

    l_flashStatus = STATUS_ERASESECTORS;
    printf("Flash CGI erase sectors.\n");
    NutSleep(100);
    l_flashError = eraseFlash(sumBytes);
    if (l_flashError != NOERROR) {
        printf("Flash CGI ERROR: error erase\n");
        l_lockFlashCgi = 0;
        return outCgi(stream, req);
    }
    l_flashStatus = STATUS_WRITEIMAGE;
    printf("Flash CGI write image.\n");
    NutSleep(100);

    l_flashError = readWriteSectorsFromStream(stream, sumBytes);
    l_lockFlashCgi = 0;
    NutSleep(100);
    if (l_flashError != NOERROR) {
        printf("\nFlash CGI ERROR image written.\n");
        return outCgi(stream, req);
    }
    l_flashStatus = STATUS_IMAGEWRITTEN;
    printf("\nFlash CGI image written.\n");

    return outCgi(stream, req);
}

